/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source
 * code directory.
 */
#include "runtime.h"
#include "appconf.h"
#include "str_hash.h"
#include "list.h"
#include "stack.h"
#include "internal.h"
#include "module.h"
#include "archdep.h"


MODULE_NAME("appconf");
MODULE_INIT(appconf_init);
MODULE_DESTROY(appconf_unload);
MODULE_DEPENDS();
MODULE_REGISTER(appconf);

int conf_foreach_callback(struct str_hash_table *t, const char *name,
		struct conf_value *cv, void *d);


struct conf_group *conf_root_group;

POOL conf_pool;
DEBUG conf_debug;

static int conf_ran_init = 0;


int appconf_load_config(const char *filename)
{
	static char filetmpbuf[1024];

	/* Try ~/.ircfs2/file */
	snprintf(filetmpbuf,1023,"%s/.%s/%s",
			getenv("HOME"),
			PACKAGE_NAME,
			filename);
	filetmpbuf[1023]=0;
	if (arch_file_exists(filetmpbuf)) 
		return appconf_loadfile(filetmpbuf);
	errno = 0;

	/* Try /etc/ircfs2/<filename> */
	snprintf(filetmpbuf,1023,"%s/%s/%s",
			SYSCONFDIR,
			PACKAGE_NAME,
			filename);
	filetmpbuf[1023]=0;
	if (arch_file_exists(filetmpbuf))
		return appconf_loadfile(filetmpbuf);
	errno = 0;

	/* Try cwd */
	snprintf(filetmpbuf,1023,"%s",
			filename);
	filetmpbuf[1023]=0;
	if (arch_file_exists(filetmpbuf))
		return appconf_loadfile(filetmpbuf);
	errno = 0;

	error(conf_debug,"Failed to find config file %s",filename);
	return -1;
}

int appconf_init()
{
	char cfgfile[255];
	if (conf_ran_init)
		return 0;
	conf_ran_init = 1;
	conf_pool = pool_new(NULL);
	conf_debug = debug_register("config");	
	debug_filter(conf_debug,D_INFO);
	conf_root_group=conf_create_group(NULL,"");

	snprintf(cfgfile,254,"%s.conf",PACKAGE_NAME);
	cfgfile[254]=0;
	if (appconf_load_config(cfgfile) == -1) {
		warn(conf_debug,"Failed to load config file %s",
				cfgfile);
	}
	/* After loading the config file, get the debug module to check
	 * its options
	 */
	debug_load_config_opts();

	return 0;
}
int appconf_unload()
{
	return -1;
}
int conf_check_init()
{
	if (conf_ran_init)
		return 0;
	appconf_init();
	error(conf_debug,"Hadn't run init!");
	return 0;
}
struct conf_value *conf_get_value(POOL p, const char *valuepath, ...)
{
	va_list vl;
	struct conf_value *cv;

	va_start(vl,valuepath);
	cv = conf_get_value_v(p,valuepath,vl);
	va_end(vl);

	return cv;
}

struct conf_value *conf_get_value_v(POOL p, const char *valuepath, va_list vl)
{
	static char tmpbuf[1024];
	vsnprintf(tmpbuf,1024,valuepath,vl);
	return conf_get_value_str(p,tmpbuf);
}

struct conf_value *conf_get_value_str(POOL p, const char *valuepath)
{
	static char tmpbuf[1024];
	char *name,*cp;
	struct conf_group *cg, *newcg;
	struct conf_value *cv;
	strncpy(tmpbuf,valuepath,1023);
	tmpbuf[1023] = 0;
	/* Split the value path into each dir */
	cp = tmpbuf;
	cg = conf_root_group;
	ASSERT(cg);
	while(*cp) {
		name = cp;
		while(*cp && *cp != '/')
			cp++;
		if (!*cp) /* Found name */
			break;
		*cp = 0;
		cp++;
		if (*name == 0)
			continue;
		cv = str_hash_get(cg->settings,name);
		if (!cv) {
			debug(conf_debug,"can't find %s in %s",
					name,cg->name);
			return NULL;
		}
		newcg = conf_value_to_group(cv);
		if (!newcg) {
			debug(conf_debug,"conf value %s in %s isn't a group",
					name,cg->name);
			return NULL;
		}
		cg = newcg;
	}
	/* Okay, name is the name of the ci in cg */
	cv = str_hash_get(cg->settings,name);
	if (cv == NULL) {
		debug(conf_debug,"Can't find item %s in group %s",
				name,cg->name);
		return NULL;
	}

	return conf_value_dup(p,cv);
}

struct conf_value *conf_group_value_n(POOL p,struct conf_group *cg, 
		int num)
{
	struct conf_value *cv;
	cv = str_hash_get_n(cg->settings,num);
	if (cv)
		return conf_value_dup(p,cv);
	return NULL;
}


struct conf_value *conf_group_value(POOL p,struct conf_group *cg, 
		const char *name)
{
	struct conf_value *cv;
	cv = str_hash_get(cg->settings,name);
	if (cv)
		return conf_value_dup(p,cv);
	return NULL;
}

const char *conf_group_name(struct conf_group *cg)
{
	return cg->name;
}



struct conf_foreach_data {
	void *appdata;
	struct conf_group *group;
	conf_foreach_func func;
};

int conf_foreach(struct conf_group *cg, conf_foreach_func func,
		void *appdata)
{
	struct conf_foreach_data fd;
	fd.func = func;
	fd.appdata = appdata;
	fd.group = cg;
	return str_hash_foreach(cg->settings,
			(str_hash_callback_t)conf_foreach_callback,
			&fd);
}
int conf_foreach_callback(struct str_hash_table *t, const char *name,
		struct conf_value *cv, void *d)
{
	struct conf_foreach_data *fd = (struct conf_foreach_data *)d;

	return fd->func(fd->group,name,cv,fd->appdata);
}

struct conf_group *conf_open_group(struct conf_group *parent,
		const char *name)
{
	struct conf_value *cv;
	struct conf_group *cg;
	if (!parent) {
		return conf_create_group(parent,name);
	}
	cv = str_hash_get(parent->settings,name);
	if (!cv) {
		return conf_create_group(parent,name);
	}
	cg = conf_value_to_group(cv);
	if (!cg) {
		error(conf_debug,"conf value %s in %s isn't a group",
				name,parent->name);
		return NULL;
	}
	return cg;
}

struct conf_group *conf_create_group(struct conf_group *parent,
		const char *name)
{
	struct conf_group *cg;
	
	cg = XMALLOC(conf_pool, struct conf_group);
	cg->name = pstrdup(conf_pool, name);
	cg->settings = str_hash_create(conf_pool);

	if (parent == NULL) {
		conf_root_group = cg;
	}
	return cg;
}

int conf_print_settings()
{
	struct conf_group *cg;
	static int inc=0;
	conf_check_init();
	
	cg = conf_root_group;
	return conf_print_group(NULL,cg,&inc);
}

int conf_print_group(struct list *l, struct conf_group *cg, void *d)
{
	int r;
	int *inc = (int *)d;
	debug(conf_debug,"%*sGroup: %s", *inc,"",cg->name);
	(*inc)++;
	r = str_hash_foreach(cg->settings, 
			(str_hash_callback_t)conf_print_hash_value,
			d);
	(*inc)--;
	debug(conf_debug,"%*sEnd of Group: %s", *inc, "", cg->name);
	return r;
}

struct conf_group *conf_value_to_group(struct conf_value *cv)
{
	if (cv->type == conf_type_group)
		return (struct conf_group *) cv->data;
	return NULL;
}

int conf_print_hash_value(struct str_hash_table *t, const char *name,
		struct conf_value *cv, void *d)
{
	int *inc = (int *) d;
	if (cv->type == conf_type_group) {
		conf_print_group(NULL,(struct conf_group *)cv->data,inc);
	} else {
		debug(conf_debug,"%*s%s[%s]: %s",
				*inc,"",name, conf_type_to_str(cv->type),
				conf_value_to_str_const(cv),*inc);
	}
	return 0;
}


const char *conf_type_to_str(enum conf_type type)
{
	static char tmpbuf[21];
	switch(type) {
		case conf_type_str: return "string";
		case conf_type_int: return "integer";
		case conf_type_float: return "float";
		case conf_type_bool: return "boolean";
		case conf_type_any: return "any";
		case conf_type_range: return "range";
		case conf_type_enum: return "constant";
		case conf_type_group: return "group";
		case conf_type_strarray: return "strarray";
		default: 
			snprintf(tmpbuf,20,"<unknown:%d>",
					type);
			tmpbuf[20] = 0;
			return tmpbuf;
	}
	return "<unknown>";
}				     

struct conf_value *conf_value_to_type(POOL p, struct conf_value *cv, 
		enum conf_type type)
{
	const char *cp;
	char **a;
	int i;
	double f;
	struct conf_group *cg;

	switch(type) {
	case conf_type_int:
		if (conf_value_to_int(cv,&i) == -1)
			return NULL;
		return conf_new_int(p,i);
	case conf_type_float:
		if (conf_value_to_float(cv,&f) == -1)
			return NULL;
		return conf_new_float(p,f);
	case conf_type_bool:
		if (conf_value_to_bool(cv,&i) == -1)
			return NULL;
		return conf_new_bool(p,i);
	case conf_type_group:
		cg = (struct conf_group *)cv->data;
		return conf_new_group_val(p,cg);
	case conf_type_strarray:
		if (conf_value_to_strarray(p,cv,&a) == -1)
			return NULL;
		return conf_new_strarray(p,a);
	case conf_type_enum: /* Enums stay as strings */
	case conf_type_str:
	default:
		cp = conf_value_to_str_const(cv);
		return conf_new_str(p,cp);
	}
}

int conf_value_to_strarray(POOL p,struct conf_value *cv, char ***ans)
{
	int i,size;
	char **a;
	const char *str;
	char **na;
	if (cv == NULL)
		return -1;
	switch(cv->type) {
	case conf_type_str:
		/* Split the string into an array and return */
		str = (char *) cv->data;
		na = strarray_create(p,str,',');
		if (na == NULL)
			return -1;
		*ans = na;
		return 0;
	case conf_type_strarray:
		/* Clone the string array */
		a = (char **)cv->data;
		for(size=0;a[size] != NULL;size++) {}
	
		na = (char **) palloc(p, (sizeof(char *) * (size+1)));
		for(i=0;a[i] != NULL; i++) {
			na[i] = pstrdup(p, a[i]);
		}

		*ans = na;
		return 0;
	default:
		/* Split the string into an array and return */
		str = conf_value_to_str_const(cv);
		if (str == NULL)
			return -1;
		na = strarray_create(p,str,',');
		if (na == NULL)
			return -1;
		*ans = na;
		return 0;
	}
}
	
int conf_value_to_str(struct conf_value *cv, char *buffer, size_t size)
{
	char *cp;
	char **a;
	int *ip;
	double *dp;
	struct conf_group *cg;
	conf_check_init();
	if (cv == NULL)
		return -1;
	switch(cv->type) {
	case conf_type_str:
		cp = (char *)cv->data;
		break;
	case conf_type_int:
		ip = (int *)cv->data;
		snprintf(buffer,size,"%d",*ip);
		return 0;
	case conf_type_float:
		dp = (double *)cv->data;
		snprintf(buffer,size,"%f",*dp);
		return 0;
	case conf_type_bool:
		ip = (int *)cv->data;
		cp = ip ? "true" : "false";
		break;
	case conf_type_strarray:
		a = (char **)cv->data;
		strarray_join(buffer,size,a,',');
		return 0;
	case conf_type_group:
		cg = (struct conf_group *) cv->data;
		snprintf(buffer,size,"group:%s",
					cg->name);
		return 0;			
	default:
		return -1;
	}
	strncpy(buffer,cp,size);	
	return 0;
}

void conf_debug_strarray(DEBUG h, char **strarray)
{
	int i;
	for(i = 0; strarray[i] != NULL; i++) {
		debug(h,"  %s",strarray[i]);
	}
}
const char *conf_value_to_str_const(struct conf_value *cv)
{
	char *cp;
	char **a;
	int *ip;
	double *dp;
	struct conf_group *cg;
	static char tmpbuf[256];
	conf_check_init();
	if (cv == NULL)
		return NULL;
	switch(cv->type) {
	case conf_type_str:
		cp = (char *)cv->data;
		return cp;
	case conf_type_int:
		ip = (int *)cv->data;
		snprintf(tmpbuf,sizeof(tmpbuf)-1,"%d",*ip);
		return tmpbuf;
	case conf_type_float:
		dp = (double *)cv->data;
		snprintf(tmpbuf,sizeof(tmpbuf)-1,"%f",*dp);
		return tmpbuf;
	case conf_type_bool:
		ip = (int *)cv->data;
		return ip ? "true" : "false";
	case conf_type_strarray:
		a = (char **)cv->data;
		strarray_join(tmpbuf,sizeof(tmpbuf),a,',');
		return tmpbuf;
	case conf_type_group:
		cg = (struct conf_group *) cv->data;
		snprintf(tmpbuf,sizeof(tmpbuf-1),"group:%s",
					cg->name);
		return tmpbuf;			
	default:
		return NULL;
	}
		
	return NULL;
}
int conf_value_to_bool(struct conf_value *cv,int *ans)
{
	char *cp;
	int *ip;
	double *dp;
	conf_check_init();
	if (cv == NULL)
		return -1;

	switch(cv->type) {
	case conf_type_str:
		cp = (char *)cv->data;
		switch(*cp) {
		case 't': case 'T': case 'y': case 'Y': case '1':
			*ans = 1;
			return 0;
		case 'f': case 'F': case 'n': case 'N': case '0':
			*ans = 0;
			return 0;
		default:
			return -1;
		}
		return -1;
	case conf_type_int:
		ip = (int *)cv->data;
		*ans = (*ip != 0) ? 1 : 0;
		return 0;
	case conf_type_float:
		dp = (double *)cv->data;
		*ans = (*dp != 0) ? 1 : 0;
		return 0;
	case conf_type_bool:
		ip = (int *)cv->data;
		*ans = *ip;
		return 0;
	default:
		return -1;
	}
	return -1;
}
		
int conf_value_to_enum(struct conf_value *cv,int *ans,
		struct conf_enum_list *el)
{
	char *cp;
	int *ip,i;
	conf_check_init();
	if (cv == NULL)
		return -1;
	switch(cv->type) {
	case conf_type_str:
	case conf_type_enum:
		cp = (char *)cv->data;
		for(i = 0; el[i].str != NULL; i++) {
			if (strcasecmp(el[i].str,cp) == 0) {
				*ans = el[i].num;
				return 0;
			}
		}
		return -1;
	case conf_type_int:
		ip = (int *) cv->data;
		/* Verify that that number exists */
		for(i = 0; el[i].str != NULL; i++) {
			if (el[i].num == *ip) {
				*ans = el[i].num;
				return 0;
			}
		}
		return -1;
	default:
		return -1;
	}

	return -1;
}
	
int conf_value_to_int(struct conf_value *cv,int *ans)
{
	char *cp;
	int *ip;
	double *dp;
	conf_check_init();
	if (cv == NULL)
		return -1;

	switch(cv->type) {
	case conf_type_str:
		cp = (char *)cv->data;
		if (!isdigit(*cp) && *cp != '-')
			return -1;
		if (*cp == '-' && !isdigit(*(cp+1)))
			return -1;
		*ans = strtol(cp,NULL,10);
		return 0;
	case conf_type_int:
		ip = (int *)cv->data;
		*ans = *ip;
		return 0;
	case conf_type_float:
		dp = (double *)cv->data;
		*ans = *dp;
		return 0;
	case conf_type_bool:
		ip = (int *)cv->data;
		*ans = *ip;
		return 0;
	default:
		return -1;
	}

	return -1;
}
		
int conf_value_to_float(struct conf_value *cv,double *ans)
{
	char *cp;
	int *ip;
	double *dp;
	conf_check_init();
	if (cv == NULL)
		return -1;

	switch(cv->type) {
	case conf_type_str:
		cp = (char *)cv->data;
		if (!isdigit(*cp) && *cp != '-')
			return -1;
		if (*cp == '-' && !isdigit(*(cp+1)))
			return -1;
		*ans = strtod(cp,NULL);
		return 0;
	case conf_type_int:
		ip = (int *)cv->data;
		*ans = *ip;
		return 0;
	case conf_type_float:
		dp = (double *)cv->data;
		*ans = *dp;
		return 0;
	case conf_type_bool:
		ip = (int *)cv->data;
		*ans = *ip;
		return 0;
	default:
		return -1;
	}
	return -1;
}
		

struct conf_value *conf_new_group_val(POOL p,struct conf_group *cg)
{
	struct conf_value *cv;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_group;
	cv->data = cg;
	return cv;
}


struct conf_value *conf_new_enum(POOL p,int i)
{
	struct conf_value *cv;
	int *ip;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_int;
	ip = XMALLOC(p, int);
	*ip = i;
	cv->data = &ip;
	return cv;
}

struct conf_value *conf_new_str(POOL p,const char *cp)
{
	struct conf_value *cv;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_str;
	cv->data = pstrdup(p,cp);
	return cv;
}


struct conf_value *conf_new_int(POOL p,int i)
{
	struct conf_value *cv;
	int *ip;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_int;
	ip = XMALLOC(p, int);
	*ip = i;
	cv->data = &ip;
	return cv;
}
struct conf_value *conf_new_bool(POOL p,int i)
{
	struct conf_value *cv;
	int *ip;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_bool;
	ip = XMALLOC(p, int);
	*ip = i;
	cv->data = &ip;
	return cv;
}


struct conf_value *conf_new_strarray(POOL p, char **a)
{
	struct conf_value *cv;
	char **na;
	int size,i;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_strarray;

	for(size=0;a[size] != NULL;size++) {}
	
	na = (char **) palloc(p, (sizeof(char *) * (size+1)));
	for(i=0;a[i] != NULL; i++) {
		na[i] = pstrdup(p, a[i]);
	}
	
	cv->data = na;
	return cv;
}

struct conf_value *conf_new_float(POOL p,double d)
{
	struct conf_value *cv;
	double *dp;
	conf_check_init();
	cv = XMALLOC(p, struct conf_value);
	cv->type = conf_type_float;
	dp = XMALLOC(p, double);
	*dp = d;
	cv->data = &dp;
	return cv;
}

struct conf_value *conf_value_dup(POOL p, struct conf_value *cv)
{
	struct conf_value *newcv;
	ASSERT(cv);
	if (p == NULL)
		p = tmp_pool;
	newcv = conf_value_to_type(p, cv, cv->type); 
	ASSERT(newcv);
	return newcv;
}

void conf_missing(DEBUG h,const char *path)
{
	error(h,"Missing the entry %s in the configuration file",path);
}
void conf_invalid(DEBUG h,const char *path, enum conf_type type)
{
	error(h,"The configuration entry %s is invalid. Expected type: %s",
			path,
			conf_type_to_str(type));
}


int parse_enum_list(const char *str,struct conf_enum_list *el, int *result)
{
	int i;
	for(i = 0; el[i].str != NULL; i++) {
		if (strcasecmp(el[i].str,str) == 0) {
			*result = el[i].num;
			return 0;
		}
	}
	return -1;
}
